#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <errno.h>
#include "cx88_mpeg_ioctl.h"
#include "tuner_devnode_device.h"
#include "dtt7612.h"
#include "or51132.h"
#include "lgh064f.h"
#include "lg3303.h"
#include "mt2131.h"
#include "cx24227.h"
#include "xc5000.h"
#include "s5h1411.h"

extern char *optarg;

void capture(uint32_t id, int fd, void *bufspace, cx88_mpeg_config &config, uint32_t time_s)
{
   cx88_mpeg_capture capture;
   capture.in_start_byte = CX88_MPEG_DEFAULT_START_BYTE;
   capture.in_start_byte_count = CX88_MPEG_DEFAULT_START_BYTE_COUNT;
   switch (id)
   {
      case 0x70633000:
         capture.in_options = 0;
         break;
      case 0x70635500:
         capture.in_options = CX88_MPEG_SERIAL_INPUT | CX88_MPEG_BYTEWIDE_SOP;
         break;
      case 0x00707801:
      case 0x18ACD618:
         capture.in_options = CX88_MPEG_SERIAL_INPUT | CX88_MPEG_PUNCTURED_CLOCK;
         break;
   }
   struct in_addr iaddr;
   struct sockaddr_in dest;
   socklen_t sz = 0;
   int outfile = 0;
   size_t msglen = (1500 / config.in_packet_size) * config.in_packet_size;
   if (!inet_aton("127.0.0.1", &iaddr))
   {
      printf("Error creating address\n");
      return;
   }
   if ((outfile = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
   {
      printf("Error opening socket\n");
      return;
   }
   sz = sizeof(dest);
   if (getsockname(outfile, (struct sockaddr *)&dest, &sz) == -1)
   {
      printf("Error getting socket name\n");
      return;
   }
   dest.sin_port = ntohs(8802);
   dest.sin_addr = iaddr;
   if (connect(outfile, (struct sockaddr *)&dest, sz) == -1)
   {
      printf("Error connecting socket\n");
      return;
   }
   /*int outfile = open("out.ts", O_CREAT | O_APPEND | O_TRUNC | O_RDWR, 0777);
   if (outfile < 0)
   {
      printf("Error opening TS output file\n");
      return;  
   }*/
   time_t start_time = time(NULL);
   printf("Beginning capture\n");
   int error = ioctl(fd, CX88_MPEG_CAPTURE, &capture);
   if (error)
   {
      printf("Error during capture ioctl\n");
      return;
   }
   cx88_mpeg_wait wait;
   wait.in_timeout_s = 0;
   do
   {
      error = ioctl(fd, CX88_MPEG_WAIT, &wait);
      if (error)
      {
         printf("Error during wait ioctl\n");
         close(outfile);
         ioctl(fd, CX88_MPEG_STOP);
         return;
      }
      if ((wait.out_status & CX88_MPEG_BUS_ERROR) == CX88_MPEG_BUS_ERROR)
      {
         printf("Fatal error: PCI bus error\n");
      }
      if ((wait.out_status & CX88_MPEG_DMA_PROGRAM) == CX88_MPEG_DMA_PROGRAM)
      {
         printf("Fatal error: DMA program error\n");
      }
      if ((wait.out_status & CX88_MPEG_WAIT_ABORTED) == CX88_MPEG_WAIT_ABORTED)
      {
         printf("Fatal error: Wait aborted\n");
      }
      if ((wait.out_status & CX88_MPEG_CAP_TIMEOUT) == CX88_MPEG_CAP_TIMEOUT)
      {
         printf("Nonfatal error: Capture timed out\n");
      }
      if ((wait.out_status & CX88_MPEG_BAD_PACKET) == CX88_MPEG_BAD_PACKET)
      {
         printf("Nonfatal error: %d bad packets\n", wait.out_num_bad_packets);
      }
      if ((wait.out_status & CX88_MPEG_HW_OVERFLOW) == CX88_MPEG_HW_OVERFLOW)
      {
         printf("Nonfatal error: %d hardware buffer overflows\n", wait.out_num_overflows);
      }
      if (wait.out_status & CX88_MPEG_CAP_FATAL)
      {
         ioctl(fd, CX88_MPEG_STOP);
         close(outfile);
         return;
      }
      if (wait.out_buffer_available == 0 || wait.out_buffer_available > config.out_num_buffers)
      {
         printf("No buffer available\n");
         ioctl(fd, CX88_MPEG_STOP);
         close(outfile);
         return;
      }
      uint32_t bufoffset, bufsize;
      if (config.in_buftype == CX88_MPEG_BUF_KERNEL)
      {
         bufoffset = (wait.out_buffer_available - 1) * config.bufconfig.out_kernelbuf.out_buffer_size;
         bufsize = config.bufconfig.out_kernelbuf.out_mmap_size - bufoffset;
      }
      else
      {
         bufoffset = (wait.out_buffer_available - 1) * (config.out_packets_per_buffer * config.in_packet_size);
         bufsize = (config.inout_num_packets * config.in_packet_size) - bufoffset;
      }
      //write(outfile, (char *)bufspace + bufoffset, (bufsize > (config.out_packets_per_buffer * config.in_packet_size) ? (config.out_packets_per_buffer * config.in_packet_size) : bufsize));
      char *bufptr = (char *)bufspace + bufoffset;
      if (bufsize > (config.out_packets_per_buffer * config.in_packet_size))
      {
         bufsize = config.out_packets_per_buffer * config.in_packet_size;
      }
      uint32_t count = 0, writecount = 0;
      while (count < bufsize)
      {
         writecount = bufsize - count;
         if (writecount > msglen)
         {
            writecount = msglen;
         }
         ssize_t sz2 = write(outfile, bufptr, writecount);
         if (sz2 < 0)
         {
            break;
         }
         bufptr += writecount;
         count += writecount;
      }
   } while ((time(NULL) - start_time) < time_s);
   ioctl(fd, CX88_MPEG_STOP);
   close(outfile);
   printf("Packets transferred: %ld\n", wait.out_packet_count);
}

typedef struct
{
   int fd;
   uint8_t mask;
}
fusion_reset_params;

int fusion_reset(fusion_reset_params *arg)
{
   cx88_common_gpio gpio;
   gpio.in_write_mask = arg->mask;
   gpio.in_read_mask = 0x0;
   gpio.inout_gpio = 0x0;
   if (ioctl(arg->fd, CX88_COMMON_GPIO, &gpio) != 0)
   {
      perror("Error during GPIO ioctl");
      return errno;
   }
   usleep(20000);
   gpio.inout_gpio = arg->mask;
   if (ioctl(arg->fd, CX88_COMMON_GPIO, &gpio) != 0)
   {
      perror("Error during GPIO ioctl");
      return errno;
   }
   usleep(20000);
   return 0;
}

int xc5000_reset(xc5000 &driver, void *context)
{
   fusion_reset_params *arg = reinterpret_cast<fusion_reset_params*>(context);
   return fusion_reset(arg);
}

void tune(cx88_common_i2c_config &i2config, uint32_t id, int fd, void *bufspace, cx88_mpeg_config &config, int time_s)
{
   int error = 0;
   dvb_channel channel;
   channel.modulation = DVB_MOD_VSB_8;
   channel.frequency_hz = 515000000;
   channel.bandwidth_hz = 6000000;
   channel.inversion = DVB_INVERSION_AUTO;
   tuner_config conf;
   error = conf.load_file("tuner_config");
   size_t num_drivers = i2config.in_num_i2c_devices;
   if (error)
   {
      printf("Error loading tuner config file\n");
      return;
   }
   tuner_devnode_device *devs[CX88_COMMON_MAX_I2C_DEVICES];
   dvb_driver *drivers[CX88_COMMON_MAX_I2C_DEVICES];
   char nodename[16];
   for (size_t i = 0; i < i2config.in_num_i2c_devices; ++i)
   {
      sprintf(nodename, "/dev/iicdev%d", i2config.devices[i].out_instance);
      if (error)
      {
         devs[i] = NULL;
      }
      else
      {
         devs[i] = new tuner_devnode_device(conf, nodename, error);
      }
      drivers[i] = NULL;
   }
   if (error)
   {
      printf("Error creating devnodes\n");
      goto cleanup;
   }
   switch (id)
   {
      case 0x70633000:
         drivers[0] = new dtt7612(conf, *(devs[0]));
         drivers[1] = new or51132(conf, *(devs[1]));
         break;
      case 0x70635500:
      {
         cx88_common_gpio gpio;
         gpio.in_write_mask = 0x1;
         gpio.in_read_mask = 0x0;
         gpio.inout_gpio = 0x0;
         if (ioctl(fd, CX88_COMMON_GPIO, &gpio) != 0)
         {
            printf("Error during GPIO ioctl\n");
            goto cleanup;
         }
         usleep(100000);
         gpio.in_write_mask = 0x1;
         gpio.in_read_mask = 0x0;
         gpio.inout_gpio = 0x1;
         if (ioctl(fd, CX88_COMMON_GPIO, &gpio) != 0)
         {
            printf("Error during GPIO ioctl 2\n");
            goto cleanup;  
         }
         usleep(200000);
         drivers[0] = new lgh064f(conf, *(devs[0]));
         drivers[1] = new lg3303(conf, *(devs[1]), DVB_IFC_NEG_POL, DVB_INPUT_SERIAL, error);
         break;
      }
      case 0x00707801:
         drivers[1] = new cx24227(conf, *(devs[1]), DVB_INPUT_SERIAL, 
            CX24227_QAM_IFREQ_44MHZ, CX24227_GPIO_ENABLE, CX24227_CLK_CONT_NINV, error);
         drivers[0] = new mt2131(conf, *(devs[0]), error);
         break;
      case 0x18ACD618:
      {
         num_drivers = 2;
         fusion_reset_params fusion_arg;
         fusion_arg.fd = fd;
         uint8_t tuner_mask = 0;
         uint32_t stream;
         if ((error = ioctl(fd, CX23885_COMMON_GET_STREAMS, &stream)) != 0)
         {
            perror("Error retrieving stream type");
            goto cleanup;
         }
         switch (stream)
         {
            case CX23885_STREAM_MPEG1:
               fusion_arg.mask = 0x2;
               tuner_mask = 0x1;
               break;
            case CX23885_STREAM_MPEG2:
               fusion_arg.mask = 0x8;
               tuner_mask = 0x4;
               break;
            default:
               error = EINVAL;
               goto cleanup;
         }
         if ((error = fusion_reset(&fusion_arg)))
         {
            goto cleanup;
         }
         fusion_arg.mask = tuner_mask;
         drivers[1] = new s5h1411(conf, *(devs[1]), *(devs[2]), DVB_INPUT_SERIAL, 
            S5H1411_IFREQ_5_38MHZ, S5H1411_IFREQ_5_38MHZ, S5H1411_GPIO_ENABLE, S5H1411_CLK_CONT_NINV, error);
         drivers[0] = new xc5000(conf, *(devs[0]), 5380000, xc5000_reset, &fusion_arg, error);
         break;
      }
   }
   for (size_t i = 0; i < num_drivers; ++i)
   {
      if (error || (drivers[i] == NULL))
      {
         printf("Error creating periphal drivers\n");
         goto cleanup;
      }
   }
   dvb_interface ifc;
   for (size_t i = 0; i < num_drivers; ++i)
   {
      error = drivers[i]->set_channel(channel, ifc);
      if (error)
      {
         printf("Error setting channel for driver %ld\n", i);
         goto cleanup;
      }
   }
   for (size_t i = 0; i < num_drivers; ++i)
   {
      error = drivers[i]->start(2000);
      if (error)
      {
         printf("Error starting driver %ld\n", i);
         goto cleanup;
      }
   }
   capture(id, fd, bufspace, config, time_s);
cleanup:
   for (size_t i = 0; i < i2config.in_num_i2c_devices; ++i)
   {
      delete drivers[i];
      delete devs[i];
   }
}

int main(int argc, char **argv)
{
   cx88_mpeg_config config;
   uint32_t id;
   uint32_t time_s = 0;
   int fd = open(argv[1], O_RDWR);
   int error = ioctl(fd, CX88_COMMON_CARD_ID, &id);
   int optch = -1;
   if (error)
   {
      printf("Error during card ID ioctl\n");
      close(fd);
      return 0;
   }
   printf("Card ID = 0x%x\n", id);
   void *buf = NULL;
   config.in_packet_size = CX88_MPEG_DEFAULT_PACKET_SIZE;
   config.in_buftype = CX88_MPEG_BUF_USER;  
   while ((optch = getopt(argc - 1, &(argv[1]), "u:k:t:")) != -1)
   {
      switch (optch)
      {
         case 't':
            time_s = strtoul(optarg, NULL, 10);
            break;
         case 'u':
         case 'k':
            if (buf != NULL)
            {
               printf("Buffer already allocated\n");
               goto cleanup;
            }
            config.inout_num_packets = strtoul(optarg, NULL, 10);
            if (optch == 'u')
            {
               config.in_buftype = CX88_MPEG_BUF_USER; 
               buf = malloc(config.inout_num_packets * config.in_packet_size);
               config.bufconfig.in_userbuf = buf;
            }
            else if (optch == 'k')
            {
               buf = MAP_FAILED;
               config.in_buftype = CX88_MPEG_BUF_KERNEL;
            }
            break;
      }
   }
   error = ioctl(fd, CX88_MPEG_CONFIG, &config);
   if (error)
   {
      printf("Error during config ioctl\n");
      goto cleanup;
   }
   printf("Actual num packets: %d\n", config.inout_num_packets);
   printf("Packets per buffer: %d\n", config.out_packets_per_buffer);
   if (config.in_buftype == CX88_MPEG_BUF_KERNEL)
   {
      buf = mmap(NULL, config.bufconfig.out_kernelbuf.out_mmap_size, PROT_READ, MAP_SHARED, fd, 0);
      if (buf == MAP_FAILED)
      {
         perror("Error mmap'ing device");
         goto cleanup;
      }
   }
   cx88_common_i2c_config i2config;
   i2config.in_bus_number = 0;
   i2config.in_num_i2c_devices = 2;
   sprintf(i2config.devices[0].in_driver, "iicdev");
   sprintf(i2config.devices[1].in_driver, "iicdev");
   switch (id)
   {
      case 0x70633000:
         i2config.devices[0].in_speed = CX88_I2C_SPEED_100KHZ;
         i2config.devices[0].in_addr = 0x61;
         i2config.devices[1].in_speed = CX88_I2C_SPEED_100KHZ;
         i2config.devices[1].in_addr = 0x15;
         break;
      case 0x70635500:
         i2config.devices[0].in_speed = CX88_I2C_SPEED_100KHZ;
         i2config.devices[0].in_addr = 0x61;
         i2config.devices[1].in_speed = CX88_I2C_SPEED_100KHZ;
         i2config.devices[1].in_addr = 0x59;
         break;
      case 0x00707801:
         i2config.devices[0].in_speed = CX88_I2C_SPEED_400KHZ;
         i2config.devices[0].in_addr = 0x61;
         i2config.devices[1].in_speed = CX88_I2C_SPEED_400KHZ;
         i2config.devices[1].in_addr = 0x19;
         break;
      case 0x18ACD618:
         {
            uint32_t stream;
            if (ioctl(fd, CX23885_COMMON_GET_STREAMS, &stream) != 0)
            {
               perror("Error retrieving stream type");
               goto cleanup;
            }
            if (stream == CX23885_STREAM_MPEG2)
            {
               i2config.in_bus_number = 1;
            }
            else if (stream != CX23885_STREAM_MPEG1)
            {
               printf("Unrecognized stream type %d\n", stream);
               goto cleanup;
            }
            i2config.in_num_i2c_devices = 3;
            sprintf(i2config.devices[2].in_driver, "iicdev");
            i2config.devices[0].in_speed = CX88_I2C_SPEED_400KHZ;
            i2config.devices[0].in_addr = 0x64;
            i2config.devices[1].in_speed = CX88_I2C_SPEED_400KHZ;
            i2config.devices[1].in_addr = 0x19;
            i2config.devices[2].in_speed = CX88_I2C_SPEED_400KHZ;
            i2config.devices[2].in_addr = 0x1A;
         }
         break;
      default:
         printf("Unrecognized card type: %d\n", id);
         goto cleanup;
   }
   error = ioctl(fd, CX88_COMMON_I2C_CONFIG, &i2config);
   if (error)
   {
      printf("Error during I2C config ioctl\n");
      goto cleanup;  
   }
   tune(i2config, id, fd, buf, config, time_s);
cleanup:
   close(fd);
   if (config.in_buftype == CX88_MPEG_BUF_KERNEL)
   {
      if (buf != MAP_FAILED)
      {
         munmap(buf, config.bufconfig.out_kernelbuf.out_mmap_size);
      }
   }
   else
   {
      free(buf);
   }
   return 0;
}

